/*
 +========================================================================+
 |         (c) 2020 - PROPERTY OF ECOLE POLYTECHNIQUE - LGPL 3.0          |
 |________________________________________________________________________|
 |   '&`   |                                                              |
 |    #    |   FILE       : overview_fem.cpp                              |
 |    #    |   VERSION    : 0.1.0                                         |
 |   _#_   |   AUTHOR(S)  : Matthieu Aussal & Marc Bakry                  |
 |  ( # )  |   CREATION   : 01.04.2020                                    |
 |  / 0 \  |   LAST MODIF : 31.10.2020                                    |
 | ( === ) |   SYNOPSIS   : Mesh, quadrature and finite elements tools    |
 |  `---'  |                                                              |
 +========================================================================+
 */

#include <castor/matrix.hpp>
#include <castor/smatrix.hpp>
#include <castor/linalg.hpp>
#include <castor/graphics.hpp>
#include "fem.hpp"

using namespace castor;

int main(int argc, char* argv[])
{
    //===============================================================
    std::cout << "+====================+" << std::endl;
    std::cout << "|    TRIANGLEDATA    |" << std::endl;
    std::cout << "+====================+" << std::endl;
    
    // Unitary triangle mesh
    matrix<double> Tvtx = {{0,0,0},{2,0,0},{0,2,0}};
    matrix<std::size_t> Ttri = {0,1,2};
    matrix<double> ctredg = {{1,1,0},{0,1,0},{1,0,0}};
    
    // Triangle data
    triangledata<double> triangle;
    triangle.update(Tvtx);
    disp(triangle.isfar(0,0,3));
    
    // Graphical representation
    figure fig;
    caxis(fig,{0,1});
    trimesh(fig,Ttri,Tvtx);
    quiver(fig,triangle.ctr(),triangle.nrm());
    quiver(fig,ctredg,triangle.tau());
    quiver(fig,ctredg,triangle.nu());
    
    //===============================================================
    std::cout << "+====================+" << std::endl;
    std::cout << "|   DOF/QUADRATURE   |" << std::endl;
    std::cout << "+====================+" << std::endl;

    // Finite element data
    femdata<double> fem(Ttri,Tvtx,lagrangeP1,3);
    disp(sum(fem.wgt()) - 2);
    disp(fem.base(0,0)+fem.base(0,1)+fem.base(0,2));
    disp(fem.grad(0,0,0)+fem.grad(0,1,0)+fem.grad(0,2,0));

    // Graphical representation
    caxis(fig,{0,1});
    std::size_t nd = size(fem.dof(),1), ng = size(fem.gss(),1);
    vermesh(fig,range(0,nd),fem.dof(),zeros(nd,1));
    vermesh(fig,range(0,ng),fem.gss(),ones(ng,1));
    quiver(fig,fem.gss(),0.2*fem.gssnrm());
        
    //===============================================================
    std::cout << "+===================+" << std::endl;
    std::cout << "|       MESHES      |" << std::endl;
    std::cout << "+===================+" << std::endl;

    // Sphere mesh
    matrix<double> X, Y, Z, Svtx, Pvtx;
    matrix<size_t> Stri, Ptri;
    std::tie(X,Y,Z) = sphere2(1000);
    X = horzcat(0,reshape(X,1,numel(X)));
    Y = horzcat(0,reshape(Y,1,numel(Y)));
    Z = horzcat(0,reshape(Z,1,numel(Z)));
    std::tie(Stri,Svtx) = tetdelaunay(X,Y,Z);
    std::tie(Stri,Svtx) = tetboundary(Stri,Svtx);

    // Graphical representation
    figure fig2;
    trimesh(fig2,Stri,Svtx);

    //===============================================================
    std::cout << "+======================+" << std::endl;
    std::cout << "| TRIANGLE INTEGRATION |" << std::endl;
    std::cout << "+======================+" << std::endl;
    
    // Exact integration VS singular
    X = 3*Svtx;
    matrix<double> ref(size(X,1),2), sol(size(X,1),2);
    double rxy, rdotn;
    for (std::size_t l=0; l<size(X,1); ++l)
    {
        // Analytic
        std::tie(ref(l,0),ref(l,1)) = triangle.intsing(X(l,0),X(l,1),X(l,2));
        
        // Numeric (gauss)
        for (std::size_t g=0; g<size(fem.gss(),1); ++g)
        {
            rxy = std::sqrt(std::pow(X(l,0)-fem.gss(g,0),2) +
                            std::pow(X(l,1)-fem.gss(g,1),2) +
                            std::pow(X(l,2)-fem.gss(g,2),2) );
            rdotn = ((X(l,0)-fem.gss(g,0))*fem.gssnrm(g,0) +
                     (X(l,1)-fem.gss(g,1))*fem.gssnrm(g,1) +
                     (X(l,2)-fem.gss(g,2))*fem.gssnrm(g,2) );
            sol(l,0) += 1./rxy * fem.wgt(g);
            sol(l,1) += rdotn/std::pow(rxy,3) * fem.wgt(g);
        }
    }
    disp(norm(ref-sol,"inf")/norm(ref,"inf"));
    
    //===============================================================
    std::cout << "+=========================+" << std::endl;
    std::cout << "| VARIATIONAL FORMULATION |" << std::endl;
    std::cout << "+=========================+" << std::endl;
        
    // Finite element space
    fem = femdata<double>(Stri,Svtx,lagrangeP1,3);
    
    // Mass matrix
    tic();
    auto Ms = mass<double>(fem);
    toc();
    disp(4*M_PI - sum(full(Ms)));
    
    // Rigidity matrix
    tic();
    auto Ks = rigidity<double>(fem);
    toc();
    disp(max(full(Ks)));
    disp(sum(full(Ks)));
    
    // Graphical representation
    figure fig3;
    spy(fig3,Ms,"b","Mass matrix");
    
    
    drawnow(fig);
    disp("done !");
    return 0;
}
