/*
 +========================================================================+
 |         (c) 2020 - PROPERTY OF ECOLE POLYTECHNIQUE - LGPL 3.0          |
 |________________________________________________________________________|
 |   '&`   |                                                              |
 |    #    |   FILE       : bem.hpp                                       |
 |    #    |   VERSION    : 0.1.0                                         |
 |   _#_   |   AUTHOR(S)  : Matthieu Aussal & Marc Bakry                  |
 |  ( # )  |   CREATION   : 01.04.2020                                    |
 |  / 0 \  |   LAST MODIF : 31.10.2020                                    |
 | ( === ) |   SYNOPSIS   : Boundary Elements Methods for standard        |
 |  `---'  |                operators S, D, Dt, H in dense format         |
 +========================================================================+
 */

#pragma once

#define CASTOR_BEM_HPP

#include <castor/matrix.hpp>
#include <castor/smatrix.hpp>
#include "fem.hpp"

namespace castor
{

//==========================================================================//
//                           ENVIRONMENT DATA                               //
//==========================================================================//
enum bemsource{PWsource, dnPWsource, SWsource, dnSWsource};


//==========================================================================//
//                              EXCITATIONS                                 //
//==========================================================================//
//==========================================================================
//
///
template<typename T>
matrix<std::complex<T>> planeWave(matrix<T>const& X, matrix<T>const& U, T wavenumber)
{
    if (size(X,2)!=size(U,2))
    {
        error(__FILE__, __LINE__, __FUNCTION__,"Vertices X and direction U should have same number of columns.");
    }
    if (norm(U,"2") != 1.)
    {
        warning(__FILE__, __LINE__, __FUNCTION__,"Plane wave direction is not normalized.");
    }
    std::complex<T> imk(0,wavenumber);
    return exp(imk*mtimes(X,transpose(U)));
}

//==========================================================================
//
///
template<typename T>
matrix<std::complex<T>> dnplaneWave(matrix<T>const& X, matrix<T>const& N, matrix<T>const& U, T wavenumber)
{
    if (size(X,2)!=size(U,2) || size(N,2)!=size(U,2))
    {
        error(__FILE__, __LINE__, __FUNCTION__,"Vertices X, normals N and direction U should have same number of columns.");
    }
    if (norm(U,"2") != 1.)
    {
        warning(__FILE__, __LINE__, __FUNCTION__,"Plane wave direction is not normalized.");
    }
    std::complex<T> imk(0,wavenumber);
    return imk * mtimes(N,transpose(U)) * exp(imk*mtimes(X,transpose(U)));
}

//==========================================================================
//
///
template<typename T, typename S>
matrix<T> rightHandSide(femdata<S>const& v, bemsource source, matrix<S>const& Y, S wavenumber=0)
{
    // Initialization
    matrix<T> B(size(v.dof(),1),1);
    matrix<T> Sxy;
    switch (source)
    {
        case PWsource:
            Sxy = planeWave<S>(v.gss(),Y,wavenumber);
            break;
        case dnPWsource:
            Sxy = dnplaneWave<S>(v.gss(),v.gssnrm(),Y,wavenumber);
            break;
        default:
            error(__FILE__, __LINE__, __FUNCTION__,"Source term not defined.");
            break;
    }
    
    // Builder
    std::size_t ig;
    for (std::size_t e=0; e<size(v.elt2dof(),1); ++e)
    {
        for (std::size_t g=0; g<size(v.elt2gss(),2); ++g)
        {
            ig = v.elt2gss(e,g);
            for (std::size_t d=0; d<size(v.elt2dof(),2); ++d)
            {
                B(v.elt2dof(e,d)) += Sxy(ig) * v.wgt(ig) * v.base(ig,d);
            }
        }
    }
    return B;
}


//==========================================================================//
//                          INTEGRAL OPERATORS                              //
//==========================================================================//
//==========================================================================
//
///
template<typename T, typename S>
matrix<T> singleLayer(matrix<S>const& X, matrix<S>const& Y, S wavenumber=0,
                      matrix<std::size_t> Ix={}, matrix<std::size_t> Iy={})
{
    // Initialize dof data
    if (numel(Ix)==0) {Ix = range(0,size(X,1));}
    if (numel(Iy)==0) {Iy = range(0,size(Y,1));}
    if (max(Ix)>size(X,1) || max(Iy)>size(Y,1))
    {
        error(__FILE__, __LINE__, __FUNCTION__,"Indices are not in data bound.");
    }
    matrix<T> M(numel(Ix),numel(Iy));
    
    // Initialize local data
    std::size_t ix, iy;
    S rxy;
    T imk=std::sqrt(T(-1))*wavenumber;
    
    // Build
    for (std::size_t i=0; i<numel(Ix); ++i)
    {
        ix = Ix(i);
        for (std::size_t j=0; j<numel(Iy); ++j)
        {
            iy  = Iy(j);
            rxy = std::sqrt((X(ix,0)-Y(iy,0))*(X(ix,0)-Y(iy,0)) +
                            (X(ix,1)-Y(iy,1))*(X(ix,1)-Y(iy,1)) +
                            (X(ix,2)-Y(iy,2))*(X(ix,2)-Y(iy,2)) );
            if (wavenumber==0)
            {
                M(i,j) = (rxy>1e-12)? (1./rxy) : (0);
            }
            else
            {
                M(i,j) = (rxy>1e-12)? (std::exp(imk*rxy)/rxy) : (imk);
            }
        }
    }
    
    // Return with 4*pi normalization
    M /= (T)(4*M_PI);
    return M;
}

//==========================================================================
//
///
template<typename T, typename S>
matrix<T> singleLayer(matrix<S>const& X, femdata<S>const& u, S wavenumber=0,
                      matrix<std::size_t> Ix={}, matrix<std::size_t> Jd={})
{
    // Initialize dof data
    if (numel(Ix)==0) {Ix = range(0,size(X,1));}
    if (numel(Jd)==0) {Jd = range(0,size(u.dof(),1));}
    if (max(Ix)>size(X,1) || max(Jd)>size(u.dof(),1))
    {
        error(__FILE__, __LINE__, __FUNCTION__,"Indices are not in data bound.");
    }
    matrix<T> M(numel(Ix),numel(Jd));
    
    // Initialize element data
    matrix<std::size_t> Je;
    matrix<long> e2du;
    std::tie(Je,e2du) = u.dofsubdata(Jd);
    
    // Initialize local data
    std::size_t ix, jg;
    matrix<S> projectP1(1,3);
    S rxy, Rm1, dnyRm1;
    T Gxy, imk=std::sqrt(T(-1))*wavenumber;
    
    // Collocation points
    for (std::size_t i=0; i<numel(Ix); ++i)
    {
        ix = Ix(i);
        // Emmiters elements
        for (std::size_t j=0; j<length(Je); ++j)
        {
            // Analytic integration
            if (u.tridata(Je(j)).isfar(X(ix,0),X(ix,1),X(ix,2))) {Rm1 = 0;}
            else
            {
                std::tie(Rm1,dnyRm1) = u.tridata(Je(j)).intsing(X(ix,0),X(ix,1),X(ix,2));
            }
            
            // Gaussian quadrature
            for (std::size_t k=0; k<size(u.elt2gss(),2); ++k)
            {
                jg = u.elt2gss(Je(j),k);
                
                // Green kernel
                rxy = std::sqrt((X(ix,0)-u.gss(jg,0))*(X(ix,0)-u.gss(jg,0)) +
                                (X(ix,1)-u.gss(jg,1))*(X(ix,1)-u.gss(jg,1)) +
                                (X(ix,2)-u.gss(jg,2))*(X(ix,2)-u.gss(jg,2)));
                if (wavenumber==0)
                {
                    Gxy = (rxy>1e-12)? (1./rxy) : (0);
                }
                else
                {
                    Gxy = (rxy>1e-12)? (std::exp(imk*rxy)/rxy) : (imk);
                }
                
                // Integration
                Gxy *= u.wgt(jg);
                
                // Dof repartition
                for (std::size_t l=0; l<size(e2du,2); ++l)
                {
                    if (e2du(j,l)!=-1)
                    {
                        M(i,e2du(j,l)) += Gxy * u.base(jg,l);
                    }
                }
                
                // Gaussian integration
                if (Rm1!=0) {Rm1 -= ((rxy>1e-12)?(1./rxy):(0)) * u.wgt(jg);}
            }
            
            // Regularization
            if (Rm1!=0)
            {
                // Lagrange P0
                if (size(e2du,2)==1) {M(i,e2du(j)) += Rm1;}
                // Lagrange P1
                else if (size(e2du,2)==3)
                {
                    u.tridata(Je(j)).projectP1(X(ix,0),X(ix,1),X(ix,2),projectP1);
                    for (std::size_t l=0; l<size(e2du,2); ++l)
                    {
                        if (e2du(j,l)!=-1)
                        {
                            M(i,e2du(j,l)) += Rm1 * projectP1(l);
                        }
                    }
                }
            }
        }
    }
    
    // Return with 4*pi normalization
    M /= (T)(4*M_PI);
    return M;
}

//==========================================================================
//
///
template<typename T, typename S>
matrix<T> singleLayer(femdata<S>const& v, femdata<S>const& u, S wavenumber=0,
                      matrix<std::size_t> Id={}, matrix<std::size_t> Jd={})
{
    // Initialize dof data
    if (numel(Id)==0) {Id = range(0,size(v.dof(),1));}
    if (numel(Jd)==0) {Jd = range(0,size(u.dof(),1));}
    matrix<T> M(numel(Id),numel(Jd));
    
    // Initialize element data
    matrix<std::size_t> Ie, Je;
    matrix<long> e2dv, e2du;
    std::tie(Ie,e2dv) = v.dofsubdata(Id);
    std::tie(Je,e2du) = u.dofsubdata(Jd);
    
    // Initialize local data
    std::size_t ig, jg;
    matrix<S> projectP1(1,3), xg(1,3);
    S rxy, Rm1, dnyRm1;
    T tmp, Gxy, imk=std::sqrt(T(-1))*wavenumber;
    
    // Receiver elements
    for (std::size_t i=0; i<length(Ie); ++i)
    {
        // Emitters elements
        for (std::size_t j=0; j<length(Je); ++j)
        {
            // Receiver gauss points
            for (std::size_t k=0; k<size(v.elt2gss(),2); ++k)
            {
                // Emmiters analytic integration for receiver gauss point
                ig = v.elt2gss(Ie(i),k);
                xg(0)=v.gss(ig,0); xg(1)=v.gss(ig,1); xg(2)=v.gss(ig,2);
                if (u.tridata(Je(j)).isfar(xg(0),xg(1),xg(2))) {Rm1 = 0;}
                else
                {
                    std::tie(Rm1,dnyRm1) = u.tridata(Je(j)).intsing(xg(0),xg(1),xg(2));
                }
                
                // Emmiter gauss points
                for (std::size_t l=0; l<size(u.elt2gss(),2); ++l)
                {
                    jg = u.elt2gss(Je(j),l);
                    
                    // Green kernel
                    rxy = std::sqrt((xg(0)-u.gss(jg,0))*(xg(0)-u.gss(jg,0))+
                                    (xg(1)-u.gss(jg,1))*(xg(1)-u.gss(jg,1))+
                                    (xg(2)-u.gss(jg,2))*(xg(2)-u.gss(jg,2)));
                    if (wavenumber==0)
                    {
                        Gxy = (rxy>1e-12)? (1./rxy) : (0);
                    }
                    else
                    {
                        Gxy = (rxy>1e-12)? (std::exp(imk*rxy)/rxy) : (imk);
                    }
                    
                    // Integration
                    Gxy *= v.wgt(ig) * u.wgt(jg);
                    
                    // Dof repartition
                    for (std::size_t m=0; m<size(e2dv,2); ++m)
                    {
                        tmp = v.base(ig,m) * Gxy;
                        for (std::size_t n=0; n<size(e2du,2); ++n)
                        {
                            if (e2dv(i,m)!=-1 && e2du(j,n)!=-1)
                            {
                                M(e2dv(i,m),e2du(j,n)) += tmp * u.base(jg,n);
                            }
                        }
                    }
                    
                    // Gaussian integration
                    if (Rm1!=0) {Rm1 -= ((rxy>1e-12)?(1./rxy):(0)) * u.wgt(jg);}
                }
                
                // Regularization
                if (Rm1!=0)
                {
                    for (std::size_t m=0; m<size(e2dv,2); ++m)
                    {
                        if (e2dv(i,m)!=-1)
                        {
                            tmp = v.base(ig,m) * v.wgt(ig) * Rm1;
                            if (size(e2du,2)==1) {M(e2dv(i,m),e2du(j)) += tmp;}
                            else if (size(e2du,2)==3)
                            {
                                u.tridata(Je(j)).projectP1(xg(0),xg(1),xg(2),projectP1);
                                for (std::size_t n=0; n<size(e2du,2); ++n)
                                {
                                    if (e2du(j,n)!=-1)
                                    {
                                        M(e2dv(i,m),e2du(j,n)) += tmp * projectP1(n);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    // Return with 4*pi normalization
    M /= (T)(4*M_PI);
    return M;
}

//==========================================================================
//
///
template<typename T, typename S>
matrix<T> singleLayerInfinite(matrix<S>const& X, femdata<S>const& u, S wavenumber=0,
                      matrix<std::size_t> Ix={}, matrix<std::size_t> Jd={})
{
    // Initialize dof data
    if (numel(Ix)==0) {Ix = range(0,size(X,1));}
    if (numel(Jd)==0) {Jd = range(0,size(u.dof(),1));}
    if (max(Ix)>size(X,1) || max(Jd)>size(u.dof(),1))
    {
        error(__FILE__, __LINE__, __FUNCTION__,"Indices are not in data bound.");
    }
    matrix<T> M(numel(Ix),numel(Jd));
    
    // Initialize element data
    matrix<std::size_t> Je;
    matrix<long> e2du;
    std::tie(Je,e2du) = u.dofsubdata(Jd);
    
    // Initialize local data
    std::size_t ix, jg;
    S xdoty;
    T Gxy, imk=std::sqrt(T(-1))*wavenumber;
    
    // Collocation points
    for (std::size_t i=0; i<numel(Ix); ++i)
    {
        ix = Ix(i);
        // Emmiters elements
        for (std::size_t j=0; j<length(Je); ++j)
        {
            // Gaussian quadrature
            for (std::size_t k=0; k<size(u.elt2gss(),2); ++k)
            {
                jg = u.elt2gss(Je(j),k);
                
                // Green kernel
                xdoty = X(ix,0)*u.gss(jg,0) + X(ix,1)*u.gss(jg,1) + X(ix,2)*u.gss(jg,2);
                Gxy   = std::exp(-imk*xdoty) * u.wgt(jg);
                
                // Dof repartition
                for (std::size_t l=0; l<size(e2du,2); ++l)
                {
                    if (e2du(j,l)!=-1)
                    {
                        M(i,e2du(j,l)) += Gxy * u.base(jg,l);
                    }
                }
            }
        }
    }
    
    // Return with 4*pi normalization
    M /= (T)(4*M_PI);
    return M;
}

//==========================================================================
//
///
template<typename T, typename S>
matrix<T> doubleLayer(matrix<S>const& X, femdata<S>const& u, S wavenumber=0,
                      matrix<std::size_t> Ix={}, matrix<std::size_t> Jd={})
{
    // Initialize dof data
    if (numel(Ix)==0) {Ix = range(0,size(X,1));}
    if (numel(Jd)==0) {Jd = range(0,size(u.dof(),1));}
    if (max(Ix)>size(X,1) || max(Jd)>size(u.dof(),1))
    {
        error(__FILE__, __LINE__, __FUNCTION__,"Indices are not in data bound.");
    }
    matrix<T> M(numel(Ix),numel(Jd));
    
    // Initialize element data
    matrix<std::size_t> Je;
    matrix<long> e2du;
    std::tie(Je,e2du) = u.dofsubdata(Jd);
    
    // Initialize local data
    std::size_t ix, jg;
    matrix<S> projectP1(1,3);
    S rxy, rdotny, Rm1, dnyRm1;
    T Gxy, imk=std::sqrt(T(-1))*wavenumber;
    
    // Collocation points
    for (std::size_t i=0; i<numel(Ix); ++i)
    {
        ix = Ix(i);
        // Emmiters elements
        for (std::size_t j=0; j<length(Je); ++j)
        {
            // Analytic integration
            if (u.tridata(Je(j)).isfar(X(ix,0),X(ix,1),X(ix,2))) {dnyRm1 = 0;}
            else
            {
                std::tie(Rm1,dnyRm1) = u.tridata(Je(j)).intsing(X(ix,0),X(ix,1),X(ix,2));
            }
            
            // Gaussian quadrature
            for (std::size_t k=0; k<size(u.elt2gss(),2); ++k)
            {
                jg = u.elt2gss(Je(j),k);
                
                // Green kernel
                rxy = std::sqrt((X(ix,0)-u.gss(jg,0))*(X(ix,0)-u.gss(jg,0)) +
                                (X(ix,1)-u.gss(jg,1))*(X(ix,1)-u.gss(jg,1)) +
                                (X(ix,2)-u.gss(jg,2))*(X(ix,2)-u.gss(jg,2)) );
                rdotny = ((X(ix,0)-u.gss(jg,0))*u.gssnrm(jg,0) +
                          (X(ix,1)-u.gss(jg,1))*u.gssnrm(jg,1) +
                          (X(ix,2)-u.gss(jg,2))*u.gssnrm(jg,2) );
                if (wavenumber==0)
                {
                    Gxy = (rxy>1e-12)? (rdotny/(rxy*rxy*rxy)) : (0);
                }
                else
                {
                    Gxy = (rxy>1e-12)? (rdotny*(1./rxy-imk)*std::exp(imk*rxy)/(rxy*rxy)) : (0);
                }
                
                // Integration
                Gxy *= u.wgt(jg);
                
                // Dof repartition
                for (std::size_t l=0; l<size(e2du,2); ++l)
                {
                    if (e2du(j,l)!=-1)
                    {
                        M(i,e2du(j,l)) += Gxy * u.base(jg,l);
                    }
                }
                
                // Gaussian integration
                if (dnyRm1!=0) {dnyRm1 -= ((rxy>1e-12)?(rdotny/(rxy*rxy*rxy)):(0)) * u.wgt(jg);}
            }
            
            // Regularization
            if (dnyRm1!=0)
            {
                // Lagrange P0
                if (size(e2du,2)==1) {M(i,e2du(j)) += dnyRm1;}
                // Lagrange P1
                else if (size(e2du,2)==3)
                {
                    u.tridata(Je(j)).projectP1(X(ix,0),X(ix,1),X(ix,2),projectP1);
                    for (std::size_t l=0; l<size(e2du,2); ++l)
                    {
                        if (e2du(j,l)!=-1)
                        {
                            M(i,e2du(j,l)) += dnyRm1 * projectP1(l);
                        }
                    }
                }
            }
        }
    }
    
    // Return with 4*pi normalization
    M /= (T)(4*M_PI);
    return M;
}

//==========================================================================
//
///
template<typename T, typename S>
matrix<T> doubleLayer(femdata<S>const& v, femdata<S>const& u, S wavenumber=0,
                      matrix<std::size_t> Id={}, matrix<std::size_t> Jd={})
{
    // Initialize dof data
    if (numel(Id)==0) {Id = range(0,size(v.dof(),1));}
    if (numel(Jd)==0) {Jd = range(0,size(u.dof(),1));}
    matrix<T> M(numel(Id),numel(Jd));
    
    // Initialize element data
    matrix<std::size_t> Ie, Je;
    matrix<long> e2dv, e2du;
    std::tie(Ie,e2dv) = v.dofsubdata(Id);
    std::tie(Je,e2du) = u.dofsubdata(Jd);
    
    // Initialize local data
    std::size_t ig, jg;
    matrix<S> projectP1(1,3), xg(1,3);
    S rxy, rdotny, Rm1, dnyRm1;
    T tmp, Gxy, imk=std::sqrt(T(-1))*wavenumber;
    
    // Receiver elements
    for (std::size_t i=0; i<length(Ie); ++i)
    {
        // Emitters elements
        for (std::size_t j=0; j<length(Je); ++j)
        {
            // Receiver gauss points
            for (std::size_t k=0; k<size(v.elt2gss(),2); ++k)
            {
                // Emmiters analytic integration for receiver gauss point
                ig = v.elt2gss(Ie(i),k);
                xg(0)=v.gss(ig,0); xg(1)=v.gss(ig,1); xg(2)=v.gss(ig,2);
                if (u.tridata(Je(j)).isfar(xg(0),xg(1),xg(2))) {dnyRm1 = 0;}
                else
                {
                    std::tie(Rm1,dnyRm1) = u.tridata(Je(j)).intsing(xg(0),xg(1),xg(2));
                }
                
                // Emmiter gauss points
                for (std::size_t l=0; l<size(u.elt2gss(),2); ++l)
                {
                    jg = u.elt2gss(Je(j),l);
                    
                    // Green kernel
                    rxy = std::sqrt((xg(0)-u.gss(jg,0))*(xg(0)-u.gss(jg,0)) +
                                    (xg(1)-u.gss(jg,1))*(xg(1)-u.gss(jg,1)) +
                                    (xg(2)-u.gss(jg,2))*(xg(2)-u.gss(jg,2)) );
                    rdotny = ((xg(0)-u.gss(jg,0))*u.gssnrm(jg,0) +
                              (xg(1)-u.gss(jg,1))*u.gssnrm(jg,1) +
                              (xg(2)-u.gss(jg,2))*u.gssnrm(jg,2) );
                    if (wavenumber==0)
                    {
                        Gxy = (rxy>1e-12)? (rdotny/(rxy*rxy*rxy)) : (0);
                    }
                    else
                    {
                        Gxy = (rxy>1e-12)? (rdotny*(1./rxy-imk)*std::exp(imk*rxy)/(rxy*rxy)) : (0);
                    }
                    
                    // Integration
                    Gxy *= v.wgt(ig) * u.wgt(jg);
                    
                    // Dof repartition
                    for (std::size_t m=0; m<size(e2dv,2); ++m)
                    {
                        tmp = v.base(ig,m) * Gxy;
                        for (std::size_t n=0; n<size(e2du,2); ++n)
                        {
                            if (e2dv(i,m)!=-1 && e2du(j,n)!=-1)
                            {
                                M(e2dv(i,m),e2du(j,n)) += tmp * u.base(jg,n);
                            }
                        }
                    }
                    
                    // Gaussian integration
                    if (dnyRm1!=0) {dnyRm1 -= ((rxy>1e-12)?(rdotny/(rxy*rxy*rxy)):(0)) * u.wgt(jg);}
                }
                
                // Regularization
                if (dnyRm1!=0)
                {
                    for (std::size_t m=0; m<size(e2dv,2); ++m)
                    {
                        if (e2dv(i,m)!=-1)
                        {
                            tmp = v.base(ig,m) * v.wgt(ig) * dnyRm1;
                            if (size(e2du,2)==1) {M(e2dv(i,m),e2du(j)) += tmp;}
                            else if (size(e2du,2)==3)
                            {
                                u.tridata(Je(j)).projectP1(xg(0),xg(1),xg(2),projectP1);
                                for (std::size_t n=0; n<size(e2du,2); ++n)
                                {
                                    if (e2du(j,n)!=-1)
                                    {
                                        M(e2dv(i,m),e2du(j,n)) += tmp * projectP1(n);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    // Return with 4*pi normalization
    M /= (T)(4*M_PI);
    return M;
}

//==========================================================================
//
///
template<typename T, typename S>
matrix<T> doubleLayerInfinite(matrix<S>const& X, femdata<S>const& u, S wavenumber=0,
                      matrix<std::size_t> Ix={}, matrix<std::size_t> Jd={})
{
    // Initialize dof data
    if (numel(Ix)==0) {Ix = range(0,size(X,1));}
    if (numel(Jd)==0) {Jd = range(0,size(u.dof(),1));}
    if (max(Ix)>size(X,1) || max(Jd)>size(u.dof(),1))
    {
        error(__FILE__, __LINE__, __FUNCTION__,"Indices are not in data bound.");
    }
    matrix<T> M(numel(Ix),numel(Jd));
    
    // Initialize element data
    matrix<std::size_t> Je;
    matrix<long> e2du;
    std::tie(Je,e2du) = u.dofsubdata(Jd);
    
    // Initialize local data
    std::size_t ix, jg;
    S xdoty, xdotny;
    T Gxy, imk=std::sqrt(T(-1))*wavenumber;
    
    // Collocation points
    for (std::size_t i=0; i<numel(Ix); ++i)
    {
        ix = Ix(i);
        // Emmiters elements
        for (std::size_t j=0; j<length(Je); ++j)
        {
            // Gaussian quadrature
            for (std::size_t k=0; k<size(u.elt2gss(),2); ++k)
            {
                jg = u.elt2gss(Je(j),k);
                
                // Green kernel
                xdoty  = X(ix,0)*u.gss(jg,0) + X(ix,1)*u.gss(jg,1) + X(ix,2)*u.gss(jg,2);
                xdotny = X(ix,0)*u.gssnrm(jg,0) + X(ix,1)*u.gssnrm(jg,1) + X(ix,2)*u.gssnrm(jg,2);
                Gxy    = (-imk*xdotny) * std::exp(-imk*xdoty) * u.wgt(jg);
                
                // Dof repartition
                for (std::size_t l=0; l<size(e2du,2); ++l)
                {
                    if (e2du(j,l)!=-1)
                    {
                        M(i,e2du(j,l)) += Gxy * u.base(jg,l);
                    }
                }
            }
        }
    }
    
    // Return with 4*pi normalization
    M /= (T)(4*M_PI);
    return M;
}

//==========================================================================
//
///
template<typename T, typename S>
matrix<T> hypersingular(femdata<S>const& v, femdata<S>const& u, S wavenumber=0,
                        matrix<std::size_t> Id={}, matrix<std::size_t> Jd={})
{
    // Initialize dof data
    if (numel(Id)==0) {Id = range(0,size(v.dof(),1));}
    if (numel(Jd)==0) {Jd = range(0,size(u.dof(),1));}
    matrix<T> M(numel(Id),numel(Jd));
    
    // Initialize element data
    matrix<std::size_t> Ie, Je;
    matrix<long> e2dv, e2du;
    std::tie(Ie,e2dv) = v.dofsubdata(Id);
    std::tie(Je,e2du) = u.dofsubdata(Jd);
    
    // Initialize local data
    std::size_t ig, jg;
    matrix<S> projectP1(1,3), xg(1,3);
    S rxy, Rm1, dnyRm1, k2=std::pow(wavenumber,2);
    T Gxy, k2ndotnG, imk=std::sqrt(T(-1))*wavenumber;
    
    // Receiver elements
    for (std::size_t i=0; i<length(Ie); ++i)
    {
        // Emitters elements
        for (std::size_t j=0; j<length(Je); ++j)
        {
            // Receiver gauss points
            for (std::size_t k=0; k<size(v.elt2gss(),2); ++k)
            {
                // Emmiters analytic integration for receiver gauss point
                ig = v.elt2gss(Ie(i),k);
                xg(0)=v.gss(ig,0); xg(1)=v.gss(ig,1); xg(2)=v.gss(ig,2);
                if (u.tridata(Je(j)).isfar(xg(0),xg(1),xg(2))) {Rm1 = 0;}
                else
                {
                    std::tie(Rm1,dnyRm1) = u.tridata(Je(j)).intsing(xg(0),xg(1),xg(2));
                }
                
                // Emmiter gauss points
                for (std::size_t l=0; l<size(u.elt2gss(),2); ++l)
                {
                    jg = u.elt2gss(Je(j),l);
                    
                    // Green kernel
                    rxy = std::sqrt((xg(0)-u.gss(jg,0))*(xg(0)-u.gss(jg,0))+
                                    (xg(1)-u.gss(jg,1))*(xg(1)-u.gss(jg,1))+
                                    (xg(2)-u.gss(jg,2))*(xg(2)-u.gss(jg,2)));
                    if (wavenumber==0)
                    {
                        Gxy = (rxy>1e-12)? (1./rxy) : (0);
                    }
                    else
                    {
                        Gxy = (rxy>1e-12)? (std::exp(imk*rxy)/rxy) : (imk);
                    }
                    
                    // Integration
                    Gxy *= v.wgt(ig) * u.wgt(jg);
                    k2ndotnG = Gxy * k2 * (v.gssnrm(ig,0)*u.gssnrm(jg,0) +
                                           v.gssnrm(ig,1)*u.gssnrm(jg,1) +
                                           v.gssnrm(ig,2)*u.gssnrm(jg,2) );
                    
                    // Dof repartition
                    for (std::size_t m=0; m<size(e2dv,2); ++m)
                    {
                        for (std::size_t n=0; n<size(e2du,2); ++n)
                        {
                            if (e2dv(i,m)!=-1 && e2du(j,n)!=-1)
                            {
                                M(e2dv(i,m),e2du(j,n)) += (k2ndotnG * v.base(ig,m) * u.base(jg,n) -
                                                           Gxy * (
                                                                  v.nxgrad(ig,m,0) * u.nxgrad(jg,n,0) +
                                                                  v.nxgrad(ig,m,1) * u.nxgrad(jg,n,1) +
                                                                  v.nxgrad(ig,m,2) * u.nxgrad(jg,n,2) ));
                            }
                        }
                    }
                    
                    // Gaussian integration
                    if (Rm1!=0) {Rm1 -= ((rxy>1e-12)?(1./rxy):(0)) * u.wgt(jg);}
                }
                
                // Regularization
                if (Rm1!=0)
                {
                    for (std::size_t m=0; m<size(e2dv,2); ++m)
                    {
                        if (e2dv(i,m)!=-1)
                        {
                            jg = u.elt2gss(Je(j),0);
                            k2ndotnG = v.base(ig,m) * v.wgt(ig) * Rm1 * k2 * (v.gssnrm(ig,0)*u.gssnrm(jg,0) +
                                                                              v.gssnrm(ig,1)*u.gssnrm(jg,1) +
                                                                              v.gssnrm(ig,2)*u.gssnrm(jg,2) );
                            u.tridata(Je(j)).projectP1(xg(0),xg(1),xg(2),projectP1);
                            for (std::size_t n=0; n<size(e2du,2); ++n)
                            {
                                if (e2du(j,n)!=-1)
                                {
                                    M(e2dv(i,m),e2du(j,n)) += k2ndotnG * projectP1(n) - v.wgt(ig) * Rm1 *
                                    (- v.nxgrad(ig,m,0) * u.tridata(Je(j)).tau(n,0)
                                     - v.nxgrad(ig,m,1) * u.tridata(Je(j)).tau(n,1)
                                     - v.nxgrad(ig,m,2) * u.tridata(Je(j)).tau(n,2)) / u.tridata(Je(j)).hgt(n);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    
    // Return with 4*pi normalization
    M /= (T)(4*M_PI);
    return M;
}

}
